# pyawd - Marmousi
# Tribel Pascal - pascal.tribel@ulb.be
"""
Contains the Marmousi class.
"""
import cv2
import numpy as np
import matplotlib.pyplot as plt
from pyawd._marmousi_data import _get_marmousi_data


class Marmousi:
    """
    Represents the Marmousi velocity field. The maximal resolution is (955px*955px). This is only available in 2D.
    """
    nx: int = 32
    """
    The width of the field, in pixels
    """
    def __init__(self, nx: int = 32):
        """
        Args:
            nx (int): The width of the field, in pixels
        """
        self.raw_data = _get_marmousi_data()
        self.raw_nx = self.raw_data.shape[0]
        self.nx = min(nx, self.raw_nx)
        self.data = cv2.resize(self.raw_data, (nx, nx))
        self.data = self.data / (np.max(self.data) * 10)

    def get_data(self) -> np.ndarray:
        """
        Returns:
            - self.data: the velocity field
        """
        return self.data

    def plot(self):
        """
        Plots the field
        """
        plt.imshow(self.data)
        plt.show()



# pyawd - AcousticWaveDataset
# Tribel Pascal - pascal.tribel@ulb.be
from typing import Tuple, List, Dict

import numpy as np
import devito as dvt

import torch.utils.data
import matplotlib.colors

COLORS = matplotlib.colors.TABLEAU_COLORS
dvt.configuration['log-level'] = "WARNING"


class AcousticWaveDataset(torch.utils.data.Dataset):
    """
    A Pytorch dataset containing acoustic waves propagating in the Marmousi velocity field
    """
    size: int
    """The number of samples to generate in the dataset"""
    nx: int
    """The discretisation size of the array (maximum size is currently 955)"""
    sx: float
    """The sub-scaling factor of the array (0.5 means $\\frac{1}{2}$ values are returned)"""
    dim: int
    """The number of dimensions of the simulations (2 or 3)"""
    ddt: float
    """The time step used for the Operator solving iterations"""
    dt: float
    """The time step used for storing the wave propagation step (this should be higher than ddt)"""
    ndt: int
    """The number of steps in the simulation, accessible for the interrogators"""
    t: float
    """The simulations duration"""
    nt: int
    """The number of steps in the simulations, for which the whole simulation is accessible"""
    interrogators: List[Tuple]
    """A list containing the coordinates of each interrogator"""
    interrogators_data: Dict[Tuple, List]
    """The measurements of each interrogator"""
    attenuation_factor: float
    """The attenuation factor in the acoustic wave equation"""
    max_velocities: np.ndarray
    """The maximal velocity in the idx propagation field"""
    epicenters: np.ndarray
    """The epicenter of each simulation"""
    force_delay: np.ndarray
    """The delay of apparition of the external force for each simulation"""
    amplitude_factor: np.ndarray
    """The amplitude factor to multiply the external force with"""
    data: np.ndarray
    """The simulations data"""
    keep_full_data: bool
    """Whether keeping the full simulation or only the interrogable one"""
    openmp: bool
    """Use openmp optimization"""

    def __init__(self, size: int, dx: float = 1000/128., nx: int = 128, sx: float = 1., ddt: float = 0.01,
                 dt: float = 2, t: float = 10, interrogators: List[Tuple] = None, attenuation_factor: float = 0.5,
                 keep_full_data: bool = True, openmp: bool = False):
        """
        Args:
            size (int): The number of samples to generate in the dataset
            dx (float): The discretisation rate of the array
            nx (int): The discretisation size of the array
            dim (int): The number of dimensions of the simulations (2 or 3)
            sx (float): The sub-scaling factor of the array (0.5 means 1/2 values are returned)
            ddt (float): The time step used for the Operator solving iterations
            dt (float): The time step used for storing the wave propagation step (this should be higher than ddt)
            t (float): The simulations duration
            attenuation_factor (float): The attenuation factor in the acoustic wave equation
            keep_full_data (bool): Whether keeping the full simulation or only the interrogable one
            openmp (bool): Use openmp optimization
        """
        try:
            if dt < ddt:
                raise ValueError('dt should be >= ddt')
            self.size = size
            self.dx = dx
            self.nx = nx
            self.sx = sx
            self.ddt = ddt
            self.dt = dt
            self.nt = int(t / self.dt)
            self.ndt = int(self.nt * (self.dt / self.ddt))
            self.interrogators = interrogators
            self.attenuation_factor = attenuation_factor
            self.force_delay = np.random.random(size) * (t/2)
            self.amplitude_factor = (0.5 * np.random.random(size) + 0.25) * 2
            self.data = np.array([])
            self.keep_full_data = keep_full_data
            self.openmp = openmp

        except ValueError as err:
            print(err)

    def generate_data(self):
        """
        Generates the dataset content by solving the Acoustic Wave PDE for each of the `epicenters`
        """
        pass

    def solve_pde(self, idx: int):
        """
        Solves the Acoustic Wave Equation for the idx parameters.
        Returns:
            (numpy.ndarray): A numpy array containing the solutions for the `ndt` steps
        """
        pass

    def interrogate(self, idx: int, point: Tuple) -> np.ndarray:
        """
        Args:
            idx (int): The number of the sample to interrogate
            point (Tuple): The interrogator position
        Returns:
            (numpy.ndarray): The amplitude measurements for the interrogator at coordinates `point` for the $idx^{th}$ sample
        """
        if point not in self.interrogators_data:
            print("Error: the interrogated point is not interrogable.")
            print("Available interrogable points:", list(self.interrogators_data.keys()))
        else:
            return self.interrogators_data[point][idx]

    def plot_item(self, idx: int):
        """
        Plots the simulation of the $idx^{th}$ sample
        Args:
            idx (int): The number of the sample to plot
        """
        pass

    def plot_interrogators_response(self, idx: int):
        """
        Plots the measurements taken by the interrogators for the $idx^{th}$ sample.
        Args:
            idx (int): The number of the sample to plot
        """
        pass

    def generate_video(self, idx: int, filename: str, nb_images: int):
        """
        Generates a video representing the simulation of the $idx^{th}$ sample propagation
        Arguments:
            idx (int): the number of the sample to simulate in the video
            filename (str): the name of the video output file (without extension)
                        The video will be stored in a file called `filename`.mp4
            nb_images (int): the number of frames used to generate the video. This should be an entire divider of the number
                         of points computed when applying the solving operator
        """
        pass

    def set_scaling_factor(self, sx: float):
        """
        Fixes a new scaling factor (0.5 means $\\frac{1}{2}$ values are returned). It should be <= 1.
        Args:
            sx (float): the new scaling factor
        """
        if sx <= 1.:
            self.sx = sx
        else:
            print("The scaling factor should be lower or equal to 1.")

    def __len__(self):
        """
        Returns:
            (int): The number of simulations in the dataset
        """
        return self.size

    def __getitem__(self, idx):
        """
        Returns:
            (Tuple): The epicenter, the simulation of the `idx`th sample, the maximal speed of propagation of the
             propagation field, the delay before the external force application and the force amplitude factor
        """
        pass


# pyawd - utils
# Tribel Pascal - pascal.tribel@ulb.be
"""
Contains unclassable functions used across the package.
"""
import matplotlib.colors
import numpy as np


def get_white_cmap() -> matplotlib.colors.LinearSegmentedColormap:
    """
    Returns:
         (matplotlib.colors.LinearSegmentedColormap): A rose-white-green colormap
    """
    colors = [(1, 0, 0.7, 1), (1, 1, 1, 0.1), (0, 1, 0.7, 1)]
    return matplotlib.colors.LinearSegmentedColormap.from_list('seismic_white', colors)


def get_black_cmap() -> matplotlib.colors.LinearSegmentedColormap:
    """
    Returns:
         (matplotlib.colors.LinearSegmentedColormap): A rose-black-green colormap
    """
    colors = [(1, 0, 0.7, 1), (0, 0, 0, 0.1), (0, 1, 0.7, 1)]
    return matplotlib.colors.LinearSegmentedColormap.from_list('seismic_black', colors)


def get_ricker_wavelet(nx: int, a: float = 0.1, x0: int = 0, y0: int = 0, sigma: float = 0.075) -> np.meshgrid():
    """
    Generates a Ricker Wavelet
    Args:
        nx (int): The grid size on which the wavelet is created
        a (float): The scaling factor
        x0 (int): The center x coordinate (the grid is assumed to be centered in `(0, 0)`)
        y0 (int): The center y coordinate (the grid is assumed to be centered in `(0, 0)`)
        sigma (float): The spreading factor
    Returns:
        (numpy.meshgrid): A numpy meshgrid containing the generated Ricker Wavelet
    """
    x = np.arange(-1. - x0 / (0.5 * nx), 1. - x0 / (0.5 * nx), 2 / nx)
    y = np.arange(-1. - y0 / (0.5 * nx), 1. - y0 / (0.5 * nx), 2 / nx)
    x, y = np.meshgrid(x, y)
    return a * (2 - x ** 2 - y ** 2) * np.exp(-(x ** 2 + y ** 2) / (2 * sigma ** 2))


def create_inverse_distance_matrix(nx: int, x0: int = 0, y0: int = 0, z0: int = 0, tau: float = None, dim: int = 2) -> np.meshgrid:
    """
    Creates an $\\frac{1}{distance}$ matrix centered around `(x0, y0)`
    Args:
        nx (int): The grid size on which the wavelet is created
        x0 (int): The center x coordinate (the grid is assumed to be centered in `(0, 0)`)
        y0 (int): The center y coordinate (the grid is assumed to be centered in `(0, 0)`)
        z0 (int): The center z coordinate (the grid is assumed to be centered in `(0, 0, 0)`)
        tau (float): The distance threshold around (x0, y0) after which the distances are set to 0
        dim (int): The number of dimensions of the generated field (2 or 3)
    Returns:
        (numpy.meshgrid): A numpy meshgrid containing the generated explosive source
    """
    distance = np.array([])
    if not tau:
        tau = nx // 2
    x = np.arange(nx)
    y = np.arange(nx)
    if dim == 2:
        x, y = np.meshgrid(x, y)
        distance = np.sqrt((x - (x0 + nx // 2)) ** 2 + (y - (y0 + nx // 2)) ** 2)
        distance[x0+nx//2, y0+nx//2] = 1
    elif dim == 3:
        z = np.arange(nx)
        x, y, z = np.meshgrid(x, y, z)
        distance = np.sqrt((x - (x0 + nx // 2)) ** 2 + (y - (y0 + nx // 2)) ** 2 + (z - (z0 + nx // 2)) ** 2)
        distance[x0+nx//2, y0+nx//2, z0+nx//2] = 1
    distance[distance > tau] = 0.
    distance[distance > 0] = 1 / distance[distance > 0]
    return distance


def create_explosive_source(nx: int, x0: int = 0, y0: int = 0, z0: int = 0, tau: float = None, dim: int = 2) -> np.meshgrid:
    """
    Creates an explosive source ($\\frac{1}{distance}$ up to $\\lfloor\\frac{nx}{10}\\rfloor$) centered around `(x0, y0)`
    Args:
        nx (int): the grid size on which the wavelet is created
        x0 (int): the center x coordinate (the grid is assumed to be centered in `(0, 0)`)
        y0 (int): the center y coordinate (the grid is assumed to be centered in `(0, 0)`)
        z0 (int): the center z coordinate (the grid is assumed to be centered in `(0, 0)`)
        tau (float): the width of the explosive source
        dim (int): the number of dimensions of the generated field (2 or 3)
    Returns:
        (numpy.meshgrid): A numpy meshgrid containing the generated explosive source
    """
    res = np.array([])
    if not tau:
        tau = nx // 10
    if dim == 2:
        s_x, s_y = create_inverse_distance_matrix(nx, x0, y0, tau=tau, dim=dim), \
            create_inverse_distance_matrix(nx, x0, y0, tau=tau, dim=dim)
        s_x[:, :x0 + nx // 2] *= -1
        s_x[:, x0 + nx // 2] = 0

        s_y[:y0 + nx // 2] *= -1
        s_y[y0 + nx // 2] = 0

        res = s_x, s_y

    elif dim == 3:
        s_x, s_y, s_z = create_inverse_distance_matrix(nx, x0, y0, z0, tau, dim), \
            create_inverse_distance_matrix(nx, x0, y0, z0, tau, dim), \
            create_inverse_distance_matrix(nx, x0, y0, z0, tau, dim)
        s_x[:, :, :x0 + nx // 2] *= -1
        s_x[:, :, x0 + nx // 2] = 0

        s_y[:, :y0 + nx // 2] *= -1
        s_y[:, y0 + nx // 2] = 0

        s_z[:z0 + nx // 2] *= -1
        s_z[z0 + nx // 2] = 0

        res = s_x, s_y, s_z

    return res



# pyawd - AcousticWaveDataset
# Tribel Pascal - pascal.tribel@ulb.be
from typing import Tuple, List, Dict

import numpy as np
import devito as dvt
import matplotlib.colors
from matplotlib.colors import TABLEAU_COLORS
import torch.utils.data
from tqdm.auto import tqdm

from pyawd.GenerateVideo import generate_video
from pyawd.utils import *
from pyawd.Marmousi import *

COLORS = TABLEAU_COLORS

dvt.configuration['log-level'] = 'WARNING'


def solve_scalar_pde(grid: dvt.Grid, nx: int, ndt: int, ddt: float, epicenter: np.ndarray, velocity_model: dvt.Function) -> np.ndarray:
    """
    Solves the Acoustic Wave Equation for the input parameters
    Args:
        grid (dvt.Grid): A Devito Grid Object
        nx (int): The discretisation size of the array
        ndt (int): The number of iteration for which the result is stored
        ddt (float): The time step used for the Operator solving iterations
        epicenter (np.ndarray): The epicenter of the Ricker Wavelet at the beginning of the simulation
        velocity_model (devito.Function): The velocity field across which the wave propagates
    Returns:
        (numpy.ndarray): A numpy array containing the solutions for the `ndt` simulation steps
    """
    u = dvt.TimeFunction(name='u', grid=grid, space_order=2, save=ndt, time_order=2)
    u.data[:] = get_ricker_wavelet(nx, x0=epicenter[0], y0=epicenter[1])
    eq = dvt.Eq(u.dt2, (velocity_model ** 2) * (u.dx2 + u.dy2))
    stencil = dvt.solve(eq, u.forward)
    op = dvt.Operator(dvt.Eq(u.forward, stencil), opt='noop')
    op.apply(dt=ddt)
    return np.array(u.data)


class ScalarAcousticWaveDataset(torch.utils.data.Dataset):
    """
    A Pytorch dataset containing acoustic waves propagating in the Marmousi velocity field. This dataset can only build 2D simulations.
    """
    size: int
    """The number of samples to generate in the dataset"""
    nx: int
    """The discretisation size of the array (maximum size is currently 955)"""
    sx: float
    """The sub-scaling factor of the array (0.5 means $\\frac{1}{2}$ values are returned)"""
    ddt: float
    """The time step used for the Operator solving iterations"""
    dt: float
    """The time step used for storing the wave propagation step (this should be higher than ddt)"""
    ndt: int
    """The number of steps in the simulation, accessible for the interrogators"""
    t: float
    """The simulations duration"""
    nt: int
    """The number of steps in the simulations, for which the whole simulation is accessible"""
    interrogators: List[Tuple]
    """A list containing the coordinates of each interrogator"""
    interrogators_data: Dict[Tuple, List]
    """The measurements of each interrogator"""
    grid: dvt.Grid
    """The devito Grid on which the equation is solved"""
    velocity_model: dvt.Function
    """The propagation speed of the wave"""
    epicenters: np.ndarray
    """The epicenter of each simulation"""
    data: np.ndarray
    """The simulations data"""
    cmap: matplotlib.colors.LinearSegmentedColormap
    """The colormap used for displaying the simulations"""
    def __init__(self, size: int, nx: int = 128, sx: float = 1., ddt: float = 0.01, dt: int = 2, t: float = 10,
                 interrogators: List[Tuple] = None):
        """
        Args:
            size (int): The number of samples to generate in the dataset
            nx (int): The discretisation size of the array (maximum size is currently 955)
            sx (float): The sub-scaling factor of the array (0.5 means $\\frac{1}{2}$ values are returned)
            ddt (float): The time step used for the Operator solving iterations
            dt (float): The time step used for storing the wave propagation step (this should be higher than ddt)
            t (float): The simulations duration
            interrogators (List[Tuple]): A list containing the coordinates of each interrogator
        """
        if interrogators is None:
            interrogators = [(0, 0)]
        try:
            if dt < ddt:
                raise ValueError('dt should be >= ddt')
            self.size = size
            self.nx = min(nx, 955)
            self.sx = sx
            self.ddt = ddt
            self.dt = dt
            self.nt = int(t / self.dt)
            self.ndt = int(self.nt * (self.dt / self.ddt))
            self.interrogators = interrogators
            self.interrogators_data = None

            self.grid = dvt.Grid(shape=(self.nx, self.nx), extent=(1., 1.))
            self.velocity_model = dvt.Function(name='c', grid=self.grid)
            self.velocity_model.data[:] = Marmousi(self.nx).get_data().T

            self.epicenters = np.random.randint(-self.nx // 2, self.nx // 2, size=(self.size, 2)).reshape(
                (self.size, 2))
            self.data = None
            self.generate_data()

            self.cmap = get_black_cmap()

        except ValueError as err:
            print(err)

    def generate_data(self):
        """
        Generates the dataset content by solving the Acoustic Wave PDE for each of the `epicenters`
        """
        self.data = []
        self.interrogators_data = {interrogator: [] for interrogator in self.interrogators}
        for i in tqdm(range(self.size)):
            data = solve_scalar_pde(self.grid, self.nx, self.ndt, self.ddt, self.epicenters[i], self.velocity_model)
            self.data.append(data[::int(self.ndt / self.nt)])
            for interrogator in self.interrogators:
                self.interrogators_data[interrogator].append(
                    data[:, interrogator[0] + (self.nx // 2), -interrogator[1] + (self.nx // 2)])
        self.data = np.array(self.data)

    def interrogate(self, idx: int, point: Tuple) -> np.ndarray:
        """
        Args:
            idx (int): the number of the sample to interrogate
            point (Tuple): the interrogator position as a Tuple
        Returns:
             (np.ndarray): The amplitude measurements for the interrogator at coordinates `point` for the `idx`th sample
        """
        if point not in self.interrogators_data:
            print("Error: the interrogated point is not interrogable.")
            print("Available interrogable points:", list(self.interrogators_data.keys()))
        else:
            return self.interrogators_data[point][idx]

    def plot_item(self, idx: int):
        """
        Plots the simulation of the $idx^{th}$ sample
        Args:
            idx (int): the number of the sample to plot
        """
        colors = {}
        i = 0
        for interrogator in self.interrogators:
            colors[interrogator] = list(COLORS.values())[i]
            i += 1
        epicenter, item = self[idx]
        fig, ax = plt.subplots(ncols=self.nt, figsize=(self.nt * 3, 3))
        for i in range(self.nt):
            ax[i].imshow(self.velocity_model.data[::int(1 / self.sx), ::int(1 / self.sx)].T,
                         vmin=np.min(self.velocity_model.data[::int(1 / self.sx), ::int(1 / self.sx)]),
                         vmax=np.max(self.velocity_model.data[::int(1 / self.sx), ::int(1 / self.sx)]), cmap="gray")
            x = ax[i].imshow(item[i * (item.shape[0] // self.nt)].T,
                             vmin=-np.max(np.abs(item[i * (item.shape[0] // self.nt):])),
                             vmax=np.max(np.abs(item[i * (item.shape[0] // self.nt):])),
                             cmap=self.cmap)
            for interrogator in self.interrogators:
                ax[i].scatter(interrogator[0] + (self.nx // 2), -interrogator[1] + (self.nx // 2), marker="1",
                              color=colors[interrogator])
            ax[i].set_title("t = " + str(i * (item.shape[0] // self.nt) * self.dt) + "s")
            ax[i].axis("off")
            fig.colorbar(x)
        plt.tight_layout()
        plt.show()

    def plot_interrogators_response(self, idx: int):
        """
        Plots the measurements taken by the interrogators for the $idx^{th}$ sample.
        Args:
            idx (int): the number of the sample to plot
        """
        colors = {}
        i = 0
        for interrogator in self.interrogators:
            colors[interrogator] = list(COLORS.values())[i]
            i += 1
        for i in self.interrogators:
            plt.plot(np.arange(0, self.ndt * self.ddt, self.ddt), self.interrogate(idx, i), color=colors[i])
        plt.xlabel("time (s)")
        plt.ylabel("Amplitude")
        plt.legend([str(i) for i in self.interrogators])
        plt.title("Amplitude measurement for each interrogator")

    def generate_video(self, idx: int, filename: str, nb_images: int):
        """
        Generates a video representing the simulation of the $idx^{th}$ sample propagation
        Arguments:
            idx (int): the number of the sample to simulate in the video
            filename (str): the name of the video output file (without extension)
                        The video will be stored in a file called `filename`.mp4
            nb_images (int): the number of frames used to generate the video. This should be an entire divider of the number
                         of points computed when applying the solving operator
        """
        u = solve_scalar_pde(self.grid, self.nx, self.ndt, self.ddt, self.epicenters[idx], self.velocity_model)
        generate_video(u[::self.ndt // nb_images], self.interrogators,
                       {i: self.interrogators_data[i][idx][::self.ndt // nb_images] for i in self.interrogators},
                       filename, nx=self.nx, dt=self.ndt * self.ddt / nb_images, c=self.velocity_model, verbose=True)

    def set_scaling_factor(self, sx: float):
        """
        Fixes a new scaling factor (0.5 means $\\frac{1}{2}$ values are returned). It should be <= 1.
        Args:
            sx (float): the new scaling factor
        """
        if sx <= 1.:
            self.sx = sx
        else:
            print("The scaling factor should be lower or equal to 1.")

    def __len__(self):
        """
        Returns:
            (int): The number of simulations in the dataset
        """
        return self.size

    def __getitem__(self, idx: int):
        """
        Returns:
            (Tuple): The epicenter and the simulation of the `idx`th sample
        """
        return self.epicenters[idx], self.data[idx][:, ::int(1 / self.sx), ::int(1 / self.sx)]



# pyawd - AcousticWaveDataset
# Tribel Pascal - pascal.tribel@ulb.be
from typing import Tuple, List

import numpy as np
import devito as dvt
from matplotlib.colors import TABLEAU_COLORS

from pyawd import AcousticWaveDataset

COLORS = TABLEAU_COLORS
dvt.configuration['log-level'] = "WARNING"


class VectorAcousticWaveDataset(AcousticWaveDataset):
    """
    A Pytorch dataset containing acoustic waves propagating in the Marmousi velocity field
    """

    def __init__(self, size: int, dx: float = 1000 / 128., nx: int = 128, sx: float = 1., ddt: float = 0.01,
                 dt: float = 2, t: float = 10, interrogators: List[Tuple] = None, attenuation_factor: float = 0.5,
                 keep_full_data: bool = True, openmp: bool = False):
        """
        Args:
            size (int): The number of samples to generate in the dataset
            dx (float): The discretisation rate of the array
            nx (int): The discretisation size of the array
            sx (float): The sub-scaling factor of the array (0.5 means 1/2 values are returned)
            ddt (float): The time step used for the Operator solving iterations
            dt (float): The time step used for storing the wave propagation step (this should be higher than ddt)
            t (float): The simulations duration
            attenuation_factor (float): The attenuation factor in the acoustic wave equation
            keep_full_data (bool): Whether keeping the full simulation or only the interrogable one
            openmp (bool): Use openmp optimization
        """
        super().__init__(size, dx, nx, sx, ddt, dt, t, interrogators, attenuation_factor,
                         keep_full_data, openmp=openmp)

    def generate_data(self):
        """
        Generates the dataset content by solving the Acoustic Wave PDE for each of the `epicenters`
        """
        pass

    def solve_pde(self, idx: int):
        """
        Solves the Acoustic Wave Equation for the idx parameters.
        Returns:
            (numpy.ndarray): A numpy array containing the solutions for the `ndt` steps
        """
        pass

    def interrogate(self, idx: int, point: Tuple) -> np.ndarray:
        """
        Args:
            idx (int): The number of the sample to interrogate
            point (Tuple): The interrogator position
        Returns:
            (numpy.ndarray): The amplitude measurements for the interrogator at coordinates `point` for the $idx^{th}$ sample
        """
        if point not in self.interrogators_data:
            print("Error: the interrogated point is not interrogable.")
            print("Available interrogable points:", list(self.interrogators_data.keys()))
        else:
            return self.interrogators_data[point][idx]

    def plot_item(self, idx: int):
        """
        Plots the simulation of the $idx^{th}$ sample
        Args:
            idx (int): The number of the sample to plot
        """
        pass

    def plot_interrogators_response(self, idx: int):
        """
        Plots the measurements taken by the interrogators for the $idx^{th}$ sample.
        Args:
            idx (int): The number of the sample to plot
        """
        pass

    def generate_video(self, idx: int, filename: str, nb_images: int):
        """
        Generates a video representing the simulation of the $idx^{th}$ sample propagation
        Arguments:
            idx (int): the number of the sample to simulate in the video
            filename (str): the name of the video output file (without extension)
                        The video will be stored in a file called `filename`.mp4
            nb_images (int): the number of frames used to generate the video. This should be an entire divider of the number
                         of points computed when applying the solving operator
        """
        pass

    def set_scaling_factor(self, sx: float):
        """
        Fixes a new scaling factor (0.5 means $\\frac{1}{2}$ values are returned). It should be <= 1.
        Args:
            sx (float): the new scaling factor
        """
        if sx <= 1.:
            self.sx = sx
        else:
            print("The scaling factor should be lower or equal to 1.")

    def __len__(self):
        """
        Returns:
            (int): The number of simulations in the dataset
        """
        return self.size

    def __getitem__(self, idx):
        """
        Returns:
            (Tuple): The epicenter, the simulation of the `idx`th sample, the maximal speed of propagation of the
             propagation field, the delay before the external force application and the force amplitude factor
        """
        pass


# pyawd
# Tribel Pascal - pascal.tribel@ulb.be
r"""
# pyawd
Pyawd (standing for Pytorch Acoustic Wave Dataset) is a powerful tool for building datasets containing custom simulations of the propagation of Acoustic Wave through a given medium.
It uses the finite differences scheme (implemented in the Devito package) to solve the Acoustic Wave Equation, and offers convenient tools for the customisation of the parameters, the handling of the data, the visualisation of the simulations.

## Acoustic Wave Equation
The equation of propagation of an acoustic wave is given by $\frac{d^2u}{dt^2} = c \nabla^2 u + f(x, y)$, where
- $u(x, y)$ is the displacement field, and can be either a scalar or a vector field
- $c(x, y)$ is the wave  propagation speed
- $\nabla^2$ is the _laplacian operator_
- $f(x, y)$ is an external force applied on the system, for which the value can vary through time


## Installation
The package (along with the dependencies) is accessible via [PyPI](https://pypi.org/project/PyAWD/):

```bash
pip install pyawd
```

## Documentation
The API documentation is available [here](https://pascaltribel.github.io/pyawd/).
Basic help is provided for each class and function, and is accessible via the Python `help()` function.

## Getting started

Basic imports:
```python
import pyawd
from pyawd import ScalarAcousticWaveDataset
```

Let us generate a Dataset made of 10 simulations. Each simulation is run in a $250\times 250$ matrix. We store the field state every $2$ seconds and we run the simulation for $10$ seconds:

```python
dataset = ScalarAcousticWaveDataset(2, nx=250, dt=2, t=10)
```

Then we plot the first simulation.

```python
dataset.plot_item(0)
```

Finally, we can generate a video of this simulation. We will use $200$ frames, which yields a final rate of $20 fps$:

```python
dataset.generate_video(0, "example", 200)
```


By default, the point `(0, 0)` contains an interrogator. This means that the continuous measurement on this position (at least with a $\Delta t=dt$) can be obtained by:

```python
dataset.interrogate((0, 0))
```

## More advanced usage
Using the `VectorAcousticWaveDataset` class, you can produce simulations in 2D which are more realistic:

```python
dataset = VectorAcousticWaveDataset(2, nx=250, dt=2, interrogators=[(-10, 0), (10, 0)], t=10)
```

Especially, the `interrogate` method provides measurements along two orthogonal dimensions:

```python
dataset.plot_item(0)
dataset.plot_interrogators_response(0)
```


## Examples
Mutliple IPython notebooks are presented in the [examples](examples/) directory. If [Jupyter](https://jupyter.org) is installed, those examples can be explored by starting Jupyter:

```bash
jupyter-notebook
```

- `ScalarAcousticWavePropagation.ipynb`: an introduction to PDE solving and simulation using Devito applied on the scalar acoustic wave propagation
- `VectorAcousticWavePropagation.ipynb`: an introduction to PDE solving and simulation using Devito applied on the vector acoustic wave propagation
- `VectorAcousticWaveDataset.ipynb`: an introduction to the VectorAcousticWaveDataset possibilities
- `Marmousi.ipynb`: a visualisation of the Marmousi velocity field used in the simulations
- `Interrogators.ipynb`: an introduction to the PyAWD Interrogators usage
- `GenerateVectorAcousticWaveDataset.ipynb`: how to generate dataset using `pyawd`

## Related Works:
- https://essd.copernicus.org/preprints/essd-2023-470/
"""
from pyawd.ScalarAcousticWaveDataset import ScalarAcousticWaveDataset
from pyawd.AcousticWaveDataset import AcousticWaveDataset
from pyawd.VectorAcousticWaveDataset import VectorAcousticWaveDataset
from pyawd.VectorAcousticWaveDataset2D import VectorAcousticWaveDataset2D
from pyawd.VectorAcousticWaveDataset3D import VectorAcousticWaveDataset3D
from pyawd.Marmousi import Marmousi



# pyawd - Marmousi
# Tribel Pascal - pascal.tribel@ulb.be
"""
Gathers functions to generate videos from a given simulation.
"""
from typing import Tuple, List, Dict

import matplotlib.pyplot as plt
from matplotlib.colors import TABLEAU_COLORS
from glob import glob
import numpy as np
from subprocess import call
from os import remove
from tqdm.auto import tqdm
from devito import Function
import pyvista as pv

from pyawd.utils import get_black_cmap, get_white_cmap

COLORS = TABLEAU_COLORS


def generate_video(img: np.ndarray, interrogators: List[Tuple] = None,
                   interrogators_data: Dict[Tuple, List] = None,
                   name: str = "test", nx: int = 32, dt: float = 0.01, c: Function = None, verbose: bool = False):
    """
    Generates a video from a sequence of images, with a scalar value on each point.
    Args:
        img (numpy.ndarray): A sequence of np.arrays containing the wave state at every timestep
        interrogators (List[Tuple]): A list containing the coordinates of each interrogator, as tuples
        interrogators_data (Dict[Tuple, List]): Couples of interrogators coordinates associated with their measured data
        name (str): The name of the file to save the data to, without the `.mp4` extension
        nx (int): The width of the plane to display (it is assumed to be a squared plane)
        dt (float): The size of the timestep between two subsequent images
        c (devito.Function): A function of space representing the wave propagation speed in each spatial point
        verbose (bool): Gives information about the video generation
    """
    colors = {}
    i = 0
    if interrogators:
        for interrogator in interrogators:
            colors[interrogator] = list(COLORS.values())[i]
            i += 1
    if verbose:
        print("Generating", len(img), "images and saving to " + name + ".mp4.")
    for i in tqdm(range(len(img))):
        if interrogators:
            fig, ax = plt.subplots(ncols=2, figsize=(10, 5), gridspec_kw={'width_ratios': [1, 1]})
            if c:
                ax[0].imshow(c.data.T, vmin=np.min(c.data), vmax=np.max(c.data), cmap="gray")
            im = ax[0].imshow(img[i].T, cmap=get_black_cmap(), vmin=-np.max(np.abs(img[i:])),
                              vmax=np.max(np.abs(img[i:])))
            plt.colorbar(im, shrink=0.75, ax=ax[0])
        else:
            fig, ax = plt.subplots(figsize=(5, 5), gridspec_kw={'width_ratios': [1]})
            if c:
                ax.imshow(c.data.T, vmin=np.min(c.data), vmax=np.max(c.data), cmap="gray")
            im = ax.imshow(img[i].T, cmap=get_black_cmap(), vmin=-np.max(np.abs(img[i:])), vmax=np.max(np.abs(img[i:])))
            ax.axis('off')
            plt.colorbar(im, shrink=0.75, ax=ax)
        if interrogators:
            for interrogator in interrogators:
                ax[0].scatter(interrogator[0] + (nx // 2), -interrogator[1] + (nx // 2), marker="1",
                              color=colors[interrogator])
                ax[1].plot(np.arange(0, len(img) * dt, dt)[:i + 1], interrogators_data[interrogator][:i + 1],
                           color=colors[interrogator])
            ax[1].set_xlabel("Time")
            ax[1].set_ylabel("Amplitude")
            ax[1].legend([str(i) for i in interrogators_data])
            ax[1].set_ylim((np.min(np.array(list(interrogators_data.values()))),
                            np.max(np.array(list(interrogators_data.values())))))
            ax[0].axis('off')
        plt.title("t = " + str(dt * i)[:4] + "s")
        plt.savefig(name + "%02d.png" % i, dpi=250)
        plt.close()

    call([
        'ffmpeg', '-loglevel', 'panic', '-framerate', str(int(1 / dt)), '-i', name + '%02d.png', '-r', '32', '-pix_fmt',
        'yuv420p', name + ".mp4", '-y'
    ])
    for file_name in glob(name + "*.png"):
        remove(file_name)


def generate_quiver_video(quiver_x: np.ndarray, quiver_y: np.ndarray, interrogators: List[Tuple] = None,
                          interrogators_data: Dict[Tuple, np.ndarray] = None, name: str = "test", nx: int = 32, dt: float = 0.01,
                          c: Function = None, max_velocity: np.ndarray = 0,
                          display_velocity_model: bool = True,verbose: bool = False):
    """
    Generates a video from a sequence of images, with a vector value on each point.
    Args:
        quiver_x (numpy.ndarray): A sequence of np.arrays containing the wave x vector coordinate at every timestep
        quiver_y (numpy.ndarray): A sequence of np.arrays containing the wave y vector coordinate at every timestep
        interrogators (List[Tuple]): A list containing the coordinates of each interrogator, as tuples
        interrogators_data (Dict[Tuple, numpy.ndarray]): Couples of interrogators coordinates associated with their measured data
        name (str): The name of the file to save the data to, without the `.mp4` extension
        nx (int): The width of the plane to display (it is assumed to be a squared plane)
        dt (float): The size of the timestep between two subsequent images
        c (devito.Function): A function of space representing the wave propagation speed in each spatial point
        max_velocity (np.ndarray): The maximal speed of propagation
        display_velocity_model (bool): Display the velocity in background or not
        verbose (bool): Gives information about the video generation
    """
    if c is None:
        c = []
    colors = {}
    i = 0
    if interrogators:
        for interrogator in interrogators:
            colors[interrogator] = list(COLORS.values())[i]
            i += 1
    if verbose:
        print("Generating", len(quiver_x), "images and saving to `" + name + ".mp4`.")
    nu_x = np.max(quiver_x)
    nu_y = np.max(quiver_y)
    for i in tqdm(range(len(quiver_x))):
        fig, axes = plt.subplots(ncols=len(interrogators)+1, figsize=((len(interrogators)+1)*5, 5))
        a, b = np.meshgrid(np.arange(nx), np.arange(nx))
        if display_velocity_model:
            axes[0].imshow(c.data[:], vmin=np.min(c.data[:]), vmax=np.max(c.data[:]), cmap="gray")
            axes[0].quiver(a, b, quiver_x[i]/nu_x, -quiver_y[i]/nu_y)
        else:
            axes[0].quiver(a, b, quiver_x[i] / nu_x, quiver_y[i] / nu_y)
        for interrogator in interrogators:
            axes[0].scatter(interrogator[0] + (nx // 2), interrogator[1] + (nx // 2), marker="1",
                            color=colors[interrogator])
        axes[0].set_title("t = " + str(i * dt) + "s")
        axes[0].axis("off")
        for inter in range(len(interrogators_data)):
            key = list(interrogators_data.keys())[inter]
            for d in range(len(interrogators_data[key])):
                axes[inter+1].plot(interrogators_data[key][d][:i], linestyle=['-', '--'][d])
            axes[inter+1].set_ylim(-max(nu_x, nu_y), max(nu_x, nu_y))
        fig.suptitle("t = " + str(dt * i)[:4] + "s, velocity factor = " + str(max_velocity)[:5])
        plt.tight_layout()
        plt.savefig(name + "%02d.png" % i, dpi=250)
        plt.close()

    call([
        'ffmpeg', '-loglevel', 'panic', '-framerate', str(int(1 / dt)), '-i', name + '%02d.png', '-r', '32', '-pix_fmt',
        'yuv420p', name + ".mp4", '-y'
    ])
    for file_name in glob(name + "*.png"):
        remove(file_name)


def generate_density_video(quiver_x: np.ndarray, quiver_y: np.ndarray, quiver_z: np.ndarray,
                           interrogators: List[Tuple] = None,
                           interrogators_data: Dict[Tuple, np.ndarray] = None, name: str = "test", nx: int = 32,
                           dx: float = 1000./32, dt: float = 0.01):
    """
    Generates a video from a sequence of images, with a vector value on each point.
    Args:
        quiver_x (numpy.ndarray): A sequence of np.arrays containing the wave x vector coordinate at every timestep
        quiver_y (numpy.ndarray): A sequence of np.arrays containing the wave y vector coordinate at every timestep
        quiver_z (numpy.ndarray): A sequence of np.arrays containing the wave z vector coordinate at every timestep
        interrogators (List[Tuple]): A list containing the coordinates of each interrogator, as tuples
        interrogators_data (Dict[Tuple, numpy.ndarray]): Couples of interrogators coordinates associated with their measured data
        name (str): The name of the file to save the data to, without the `.mp4` extension
        nx (int): The width of the plane to display (it is assumed to be a squared plane)
        dx (float): The distance between each neighbour spatial point
        dt (float): The size of the timestep between two subsequent images
    """
    lengths = np.sqrt(quiver_x**2 + quiver_y**2 + quiver_z**2)*np.sign(np.mean([quiver_x, quiver_y, quiver_z], axis=0))
    interrogators_matrix = np.zeros(shape=lengths[0].shape)
    for i in interrogators:
        interrogators_matrix[i[0]+nx//2, i[1]+nx//2, i[2]+nx//2] = 1000.
    clim = [-np.max(np.abs(lengths))*0.8, np.max(np.abs(lengths))*0.8]
    p = pv.Plotter(shape=(1, 2+len(interrogators)), notebook=False, off_screen=True, window_size=[(2+len(interrogators))*512, 512])
    grid = pv.ImageData(spacing=(dx, dx, dx), origin=(nx//2, nx//2, nx//2))
    grid.dimensions = np.array(lengths[0].shape) + 1
    grid.cell_data["values"] = (lengths[0]+interrogators_matrix).flatten(order="F")
    p.add_mesh(grid, clim=clim, cmap=get_white_cmap())
    p.subplot(0, 1)
    p.add_mesh_slice_orthogonal(grid, clim=clim, cmap=get_white_cmap())
    p.link_views()
    x_range = np.arange(0, len(interrogators_data[interrogators[0]][0])*dt, dt)
    interrogators_plot = []
    for i in range(len(interrogators)):
        p.subplot(0, i+2)
        interrogators_plot.append(pv.Chart2D())
        interrogators_plot[-1].title = str(interrogators[i])
        for j in range(3):
            interrogators_plot[-1].line(x_range,
                                        interrogators_data[interrogators[i]][j],
                                        style=["-", "--", ":"][j],
                                        label=["Abscissa", "Ordinate", "Applicate"][j])
        p.add_chart(interrogators_plot[-1])
        interrogators_plot[-1].x_range = [0, dt]
    p.open_movie(name+".mp4", framerate=int(1/dt), quality=7)
    p.write_frame()
    for i in tqdm(range(len(quiver_x))):
        values = lengths[i] + interrogators_matrix
        grid.cell_data["values"] = values.flatten(order="F")
        for j in range(len(interrogators_plot)):
            interrogators_plot[j].x_range = [0, (i+1)*dt]
        p.camera.azimuth = (p.camera.azimuth+0.25)%360.
        p.write_frame()
    p.close()